/***************************** FILE HEADER *****************************
#               Copyright (c) 2000 by Columbia University              *
#                        All Rights Reserved                           *
#***********************************************************************
#***********************************************************************
#
#  File Name    : sendrecv.c
#
#  Module       : Audio Conference Server
#
#  Author(s)    : Kundan Singh (kns10@cs.columbia.edu)
#                 Gautam Nair  (gnair@cs.columbia.edu)
#
#  Created      : 28 Apr 2000
#
#  Description  : Transmit and receive threads for RTP/Audio packets.
#                
#  Notes        : 
#
#***********************************************************************
#                Modification History (Reverse Order)
#-----------------------------------------------------------------------
#   Date         Author              Description
#-----------------------------------------------------------------------
#              |             | 
#              |             | 
#-----------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>   /* select, timespec */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <math.h>       /* fabs */
#include <fcntl.h>

#include "error.h"
#include "timespec.h"
#include "list.h" 

#include <sys/stat.h>
#include <ctype.h>
#include <wchar.h>
#include <../libdb++/dbapi.h>  /* query(), initializeConnection() */
#include <mysql.h>
#include "sndfile.h"          /* recording file format*/


/*
#ifdef USE_RTP
#include "rtp_api.h"    
#include "rtp_highlevel.h"
#else
#error "Can not build this module without RTP. configure --with-rtp option"
#endif
*/

#include "rtpmapi.h"
#include "stdtypes.h"
#include "sendrecv.h"
#include "buffer.h"
#include "sdp.h"
#include "transcode.h"

/****For rtpdump*****/
#include "../librtp/rtp_highlevel_internal.h"

/* Default size for buffers */
#define BUFFER_SIZE 1500


/* payload type map */
static struct { 
       char *enc;   /*encoding name*/
       int rate;    /*sampling rate for audio; clock rate for video*/
       int ch;      /*audio channels;0 for video*/
   } pt_map[256];

extern list_t glConference; /* list of conferences */
extern list_t glUsers;      /* List of users */

static int  oSendBackAudio = 0;

static int gdPacketizationTime = 40;

/* the following for RTPdump recording */
static char fn_rtp[50];
static FILE *fp_rtp;
static struct sockaddr_in sin;
static struct timeval start;
static double dstart;
static oFileHeaderInit = 0;
#define RTP_PORT 9999


#undef INCLUDE_SSRC

//extern FILE *CreateSoundFile(char *filename, encoding_used sndformat); 
//extern void CloseSoundFile(FILE *FN); 

/* Different threads for recieving and sending RTP/RTCP packets */
/*
 * RTPMSetPacketizationTime is used to set the default packetization
 * time to application specific value.
 */
void RTPMSetPacketizationTime(int t)
{
  if ( t < 20 ) {
    t = 20;
  }
  else if ( t > 2000 ) {
    t = 2000;
  }

  gdPacketizationTime = t;
}

/*
 * This thread is used to receive packets (RTP/RTCP) for
 * a particular call leg in a conference from
 * a particular user
 */
void *RTPReceiveThread(void *param)
{
  int sock_rtp, sock_rtcp;
  int legid = (int) param;
  rtpm_call_leg_t *leg;
  rtperror the_err;
  fd_set selection;
  int maxfd = 0;
  int rtpcount = 0;
  char *theload;
  int loaduse;
  media_buffer_t *buf;
  rtp_packet rtppkt;
  rtcp_packet rtcppkt;
  int  pt = -1;
  char *ptstr = NULL;
  struct timeval initialtime, socktv;
  struct timespec next;
  long sum_delay = 0;
  long count_delay = 0;

  /*****Enhancement****/
  ResultTable table;
  int ret, i, rtpdump;
  char recording_rtp[20][1024];
  int numfmt, m, cmp;
  char recformat[3][32];
  char fmt_temp[32];
  char id[32], recordingformat[32];
  rtpm_conf_t *conf;

  /******end******/
  
  /********
  * for RTP dump declared for static outside
  char fn_rtp[50];
  FILE *fp_rtp;
  struct sockaddr_in sin;
  ***/              
  int trunc = 1000000; /* bytes to show for F_dump */
  int ff;  //flag for rtp/rtcp, 0 for rtp, 1 for rtcp
  extern void rtpdump_header(FILE *, struct sockaddr_in *, struct timeval *);
  extern double tdbl(struct timeval *);
  extern void packet_handler(FILE *, int, double, struct timeval,
                             int, struct sockaddr_in, int, RD_buffer_t);
  

  debug("RTPReceiveThread", "started for leg %d\n", legid);

  leg = ListElement(&glUsers, legid);
  if ( leg == NULL ) {
    debug("RTPReceiveThread", "No call leg %d found\n", legid);
    return NULL;
  }
  LOCK(leg->mutex);

  debug("RTPReceiveThread", "got the lock\n");

  while ( leg->rtp_context == -1 ) {
    COND_WAIT(leg->cond, leg->mutex);
    if ( leg->oTerminate ) {
      UNLOCK(leg->mutex);
      return NULL;
    }
  }

  /* Get RTP and RTCP sockets */

  the_err = RTPSessionGetRTPSocket(leg->rtp_context, 
                         &sock_rtp);
  if ( the_err != RTP_OK ) {
    error("RTPReceiveThread", "RTPSessionGetRtpSocket %s\n", 
          RTPDebugStr());
    UNLOCK(leg->mutex);
    return NULL;
  }
  else {
    debug("RTPReceiveThread", "RTP socket is %d\n", sock_rtp);
  }

  if ( maxfd < sock_rtp ) {
    maxfd = sock_rtp;
  }

  the_err = RTPSessionGetRTCPSocket(leg->rtp_context,
                                    &sock_rtcp);
  if ( the_err != RTP_OK ) {
    error("RTPReceiveThread", "RTPSessionGetRtcpSocket %s\n", 
          RTPDebugStr());
    UNLOCK(leg->mutex);
    return NULL;
  }
  else {
    debug("RTPReceiveThread", "RTCP socket is %d\n", sock_rtcp);
  }

  if ( maxfd < sock_rtcp ) {
    maxfd = sock_rtcp;
  }

  gettimeofday(&initialtime, NULL);
  gettimeofday(&start, NULL);  
 
  /* Query the recording url name from the database */

  conf = ListElement(&glConference, leg->confid);

  table = query("conferences", "url", conf->conference, 
       "id, recordingformat",
       &ret); 

  debug("RTPReceiveThread", "Rows returned= %d\n", ret);
  if (ret > 0) {
    debug("RTPReceiveThread", "Query id= %s, recordingformat= %s\n", table[0][0], table[0][1]);
  }

  strcpy(id, table[0][0]);
  strcpy(recordingformat, table[0][1]);

  m = 0; numfmt = 0;
  for (i = 0; i < strlen(recordingformat); i++) {
    if ((toascii(recordingformat[i]) >= 97) && (toascii(recordingformat[i]) <= 122)) {
      fmt_temp[m] = recordingformat[i];
      m++;
    } else {
      fmt_temp[m] = '\0';
      m = 0;
      strcpy(recformat[numfmt], fmt_temp);
      numfmt++;
    }
  }
  fmt_temp[m] = '\0';
  strcpy(recformat[numfmt], fmt_temp);
  numfmt++;

  rtpdump = 0; 
  for (i = 0; i < numfmt; i++) {
    strcpy(fmt_temp, recformat[i]);
    cmp = strcmp(fmt_temp, "rtpdump");
    if (cmp == 0) { 
      debug("RTPReceiveThread", "Found rtpdump format\n");
      rtpdump = 1;
      cmp = 1;
    }
  }
  
  /****end of query portion ****/

  /* recording using RTPdump format here */
  /* create file and initialize file header for RTPdump */

  if (rtpdump) {
    if(oFileHeaderInit == 0) {
      sprintf(fn_rtp,"conf_%s.rtp",id);
      /* open a file for binary writable */ 
      if ((fp_rtp = fopen(fn_rtp,"wb")) == NULL)
      debug("RTPReceiveThread","File %s cannot be opened\n", fn_rtp);
   
      /*initialize the address/port receiving RTP/RTCP in sipconf */
      memset(&sin,0,sizeof sin);
      sin.sin_family = AF_INET;
      sin.sin_addr.s_addr = INADDR_ANY;
      sin.sin_port = RTP_PORT;  //not sure now
 
      rtpdump_header(fp_rtp, &sin, &start);
      dstart = tdbl(&start);
      oFileHeaderInit = 1;
    }
  }  

  /* While call not closed */

  while ( !leg->oTerminate ) {

    next = timespec_delay(initialtime, 15/1000.0 ) ; /* 10 millisec */
    COND_TIMEDWAIT(leg->cond, leg->mutex, next);
    if ( leg->oTerminate ) {
      break; /* from while loop */
    }

    gettimeofday(&initialtime, NULL);

    FD_ZERO(&selection);
    FD_SET(sock_rtp, &selection);
    FD_SET(sock_rtcp, &selection);

    socktv.tv_sec = 0;
    socktv.tv_usec = 0; /* 20 millisec */

    //debug("RTPReceiveThread", "select max=%d\n", maxfd+1);

    /* Recv UDP packets from RTP/RTCP socket */
    while ( select(maxfd + 1, &selection, NULL, NULL, &socktv) > 0 ) {

      
      /* special use for rtpdump recording */
      RD_buffer_t packet;
      struct timeval now;
      double dnow;
      struct sockaddr_in *sin_client;      
      hl_context *uc;
      rtperror err;
            

      if ( FD_ISSET (sock_rtp, &selection)) {
        
	ff = 0;  /*notify recording "RTP" packets */
        //debug("RTPReceive", "Somthing on RTP socket\n");
        
        buf = NewBuffer(BUFFER_SIZE); /* XXX some size */
        if ( buf == NULL ) {
          debug("RTPReceive", "NewBuffer failed\n");
          break; //continue;
        }
        theload = buf->data;
        loaduse = buf->maxsize;

        the_err = RTPReceive(leg->rtp_context, 
                             sock_rtp, 
                             theload, 
                             &loaduse);
                  
        buf->size = loaduse;
        
        /* make sure that the packet is not sent by me */
        if ( the_err == RTP_PACKET_LOOPBACK ) {
          debug("RTPReceiveThread", "Got my own RTP packet\n");
        }
        else if ( the_err != RTP_OK) {
          debug("RTPReceiveThread", "RTPReceive %s\n", RTPDebugStr());
        }
        else {

          if (rtpdump) {
	    //here is the safe place for recording; packets not sent by us
            gettimeofday(&now, NULL);
            dnow = tdbl(&now);
            memcpy(&(packet.p.data), theload, loaduse);
            /* sin_client is socket address of client rtp */
            err = RTPSessionGetHighLevelInfo(leg->rtp_context, (void**)&uc);
            sin_client = (struct sockaddr_in *)&uc->rtp_sourceaddr;
            packet_handler(fp_rtp, trunc, dstart, now, ff, *sin_client,
                           loaduse, packet);
          } 


          rtppkt = RTPGetRTPPacket(theload, loaduse);
          rtpcount ++;
          /* First, some details about the packet */

          /*
          debug("RTPReceive",
                "%d: RTP len=%d, ts=%lu, seq=%lu\n",
                (int)rtpcount, 
                (int)loaduse,
                (unsigned long) rtppkt.RTP_header->ts,
                (unsigned short) rtppkt.RTP_header->seq);
          */

          if ( !ptstr || (int)(rtppkt.RTP_header->mpt & 0x7f) != pt) {
            pt = (int) (rtppkt.RTP_header->mpt & 0x7f);
            ptstr = GetPayloadType(&pt,
                                   &leg->sdp_rx);
            debug("GetPayloadType", "returned %s for %d\n",
                          ptstr ? ptstr : "NULL", pt);
          }
          
          if ( ptstr == NULL ) {
            debug("RTPReceiveThread",
                  "GetPayloadType returned NULL for %x\n",
                  rtppkt.RTP_header->mpt & 0x7f);
          }
          else {
            media_buffer_t *decodedbuf;
            unsigned long tj, ai;
            double ni;
            struct timeval tv;
            
            /* Transcode the payload to linear 16 */

            decodedbuf = 
              MediaTranscode(
                             ptstr,
                             "l16",
                             rtppkt.payload /* data */, 
                             (int) rtppkt.payload_len /* data length */ );
            if ( decodedbuf == NULL ) { //SSS
              debug("MediaTranscode", "Params are %s, %d, len = %d\n",
                                       ptstr ? ptstr : "NULL", (int) rtppkt.payload,
                                        (int) rtppkt.payload_len);
            } 
            else {
            
              /* Calculate the playout time for the decoded buffer XXX */
              /* Arrival time stamp, 8 per millisec for 8kHz sampling. */
              gettimeofday(&tv, NULL);
              ai = (unsigned long) (tv.tv_sec * 1000 + tv.tv_usec / 1000) * 8;
              tj = rtppkt.RTP_header->ts;
              
              if ( leg->delay.first_time ) {
                /* First packet, initialize di, vi */
                leg->delay.first_time = 0;
                leg->delay.vi = 0;
                //SSS
                leg->delay.di = (double) ((long)(ai - tj));
                leg->delay.pi = ai;
                leg->delay.ti = tj;
              }
            
              ni = (double)((long)(ai - tj));
              leg->delay.di = DELAY_ALPHA * leg->delay.di + 
                (1-DELAY_ALPHA) * ni;
              leg->delay.vi = DELAY_ALPHA * leg->delay.vi + 
                (1-DELAY_ALPHA) * fabs(leg->delay.di - ni);
              
              if ( rtppkt.RTP_header->mpt & 0x80 ) {
                /* First packet of talk spurt */
                leg->delay.ti = tj;
                leg->delay.pi = 
                  ((long) tj) + (long)(leg->delay.di + DELAY_MU * leg->delay.vi);
                decodedbuf->param = leg->delay.pi + DELAY_INITIAL;
                debug("RTPReceiveThread", "Talkspurt\n");
              }
              else {
                decodedbuf->param =
                  leg->delay.pi + ((long)(tj - leg->delay.ti))
                  + DELAY_INITIAL;
              }

              if ( decodedbuf->param < ai ) {
                decodedbuf->param = ai + 320;
		leg->delay.di += 320;
		leg->delay.first_time = 1; /* Reset BBB */
              }
              
              {
                /* Print average delay */
                sum_delay += ((long)(decodedbuf->param - ai))/8;
                count_delay ++;
                
                if ( count_delay > 100 ) {
                  long delay_avg = sum_delay / count_delay;
                  
                  debug("RTPReceiveThread", 
                        "delay avg = %ld ms\n",
                        delay_avg);

                  /* Something is wrong is delay avg is more than 
                   * 10 seconds. Delete the queue so that
                   * it does not build up due to an error */
                  if ( delay_avg >= 10000 || delay_avg < -3000 ) {
                    debug("RTPReceiveThread",
                          "Deleting queue since delay is more than 10 s\n");
                    DeleteQueue(&leg->rxbuflist);
                    leg->delay.first_time = 1; //BBB
                  }
  
                  count_delay = 0;
                  sum_delay = 0;
                }
              }
              
              //decodedbuf->param = ai + 200; // XXX 
              
              if ( ((long)decodedbuf->param) < ((long)ai)) {
                leg->delay.di += 
                  ((long)(ai - decodedbuf->param));
              }
              /* put it in the queue, or free the buffer. */
              BufferEnqueue(&leg->rxbuflist, decodedbuf);
            }
          }
        }
        
        /* Free the RTP buffer */
        FreeBuffer(buf);
      }

      if ( FD_ISSET(sock_rtcp, &selection)) {
         ff = 1;
 /* to notify recording "rtcp" packets */
        //debug("RTPReceive", "Somthing on RTCP socket\n");
        
        buf = NewBuffer(BUFFER_SIZE); /* XXX some size */
        if ( buf == NULL ) {
          debug("RTPReceive", "NewBuffer failed for RTCP\n");
          break; //continue;
        }
        theload = buf->data;
        loaduse = buf->maxsize;
        
        the_err = RTPReceive(leg->rtp_context, 
                             sock_rtcp, 
                             theload, 
                             &loaduse);
        buf->size = loaduse;

        /* RTCP packet is just ignored */
        
        /* make sure that the packet is not sent by me */
        if ( the_err == RTP_PACKET_LOOPBACK ) {
          debug("RTPReceiveThread", "Got my own RTCP packet\n");
        }
        else if ( the_err != RTP_OK) {
          debug("RTPReceiveThread", "RTPReceive RTCP %s\n",
                RTPDebugStr());
          debug("RTPReceiveThread",
                "Context %d(%d), sock %d\n", (int) (leg->rtp_context),
                legid, sock_rtcp);
        }
        else {
          debug("RTPReceiveThread", "RTPReceive RTCP %d\n",
                buf->size);

          if (rtpdump) {
            //here is the safe place for RTCP recording; packets not sent by us
            gettimeofday(&now, NULL);
            dnow = tdbl(&now);
            memcpy(&(packet.p.data), theload, loaduse);
            /* sin_client is socket address of client rtp */
            err = RTPSessionGetHighLevelInfo(leg->rtp_context, (void**)&uc);
            sin_client = (struct sockaddr_in *)&uc->rtcp_sourceaddr;
            packet_handler(fp_rtp, trunc, dstart, now, ff, *sin_client,
                           loaduse, packet);
          }         
          /* XXX process the received RTCP packet */
        }
        
        FreeBuffer(buf);
      }

      FD_ZERO(&selection);
      FD_SET(sock_rtp, &selection);
      FD_SET(sock_rtcp, &selection);
    } /* while select() succeeds */

  } /* While the thread exists */

  UNLOCK(leg->mutex);
  return NULL;
}

/*
 * This thread is for each conference for sending the
 * RTP packets to all the participants
 */
void *RTPSendThread(void *param)
{
  int confid = (int) param;
  rtperror the_err;
  rtpm_conf_t *conf;
  int *userid;
  rtpm_call_leg_t *user;
  /* 1st change */
  media_buffer_t *mixed, *raw, *encoded, *prevptr, *indiv, *safebufdata;
  FILE *fp_au, *fp_wav;
  char fn_au[50], fn_wav[50];
  encoding_used enc_au, enc_wav;
  u_int32 rec_size_au, rec_size_wav;
  long f_offset=0;
  unsigned long a1, a2, b1, b2;
  int default_duration; /* millisec */
  int i,j,k;
  int oBreak = 0;
  char *ptstr;
  int marker = 1; /* marker bit in outgoing packet */
  struct timespec next;
  struct timeval initialtime, tv;
  int outputsize = 0;
  /* List of Contributing sources */
#ifdef INCLUDE_SSRC
  unsigned int ssrc_list[16];
  unsigned int last_ssrc_list[16];
  int ssrc_count = 0;
  int last_ssrc_count = 0;
  int idx3;
#endif
  double total_duration; /* Total duration in sec */
  double duration_interval; /* Duration interval in seconds */
  long bytescount = 0;
  long timecount = 0;
  time_t actualtime = 0;

  /**********Enhancement *********/
  ResultTable table;
  int ret, wav, au;
  char recording_wav[20][1024];
  char recording_au[20][1024];
  int numfmt, m, cmp;
  char recformat[3][32];
  char fmt_temp[32];
  char id[32], recordingformat[32];
  /************end**************/

  default_duration = gdPacketizationTime;

  debug("RTPSendThread", "started for conf %d\n", confid);
  conf = ListElement(&glConference, confid);

  if ( conf == NULL ) {
    debug("RTPSendThread", "No conf %d found\n", confid);
    return NULL;
  }
  LOCK(conf->mutex);

  /* Query the recording url name from database */

  table = query("conferences", "url", conf->conference, 
       "id, recordingformat",
       &ret); 

  
  debug("RTPReceiveThread", "Rows returned= %d\n", ret);
  if (ret > 0) {
    debug("RTPReceiveThread", "Query result id= %s, recordingformat= %s\n", table[0][0], table[0][1]);
  }

  strcpy(id, table[0][0]);
  strcpy(recordingformat, table[0][1]);

  m = 0; numfmt = 0;
  for (i = 0; i < strlen(recordingformat); i++) {
    if ((toascii(recordingformat[i]) >= 97) && (toascii(recordingformat[i]) <= 122)) {
      fmt_temp[m] = recordingformat[i];
      m++;
    } else {
      fmt_temp[m] = '\0';
      m = 0;
      strcpy(recformat[numfmt], fmt_temp);
      numfmt++;
    }
  }
  fmt_temp[m] = '\0';
  strcpy(recformat[numfmt], fmt_temp);
  numfmt++;

  wav = 0; au = 0; 
  for (i = 0; i < numfmt; i++) {
    strcpy(fmt_temp, recformat[i]);
    cmp = strcmp(fmt_temp, "au");
    if (cmp == 0) { 
      debug("RTPReceiveThread", "Found au format\n");
      au = 1;
      cmp = 1;
    }
    cmp = strcmp(fmt_temp, "wav");
    if (cmp == 0) { 
      debug("RTPReceiveThread", "Found wav format\n");
      wav = 1;
      cmp = 1;
    }    
  }


  /***end of query portion  ************/

  /*since conf exists, open file to record*/

  /*for recording using .au file */
  if (au) {
    rec_size_au=0;
    enc_au.channel=1;
    enc_au.encoding=1;
    strcpy(enc_au.magic,".snd");
    enc_au.sample_rate=8000;  
    sprintf(fn_au,"conf_%s.au",id);
    if((fp_au=CreateSoundFile(fn_au,enc_au))==NULL)
      debug("RTPSendThread","File %s cannot open\n",fn_au);
  }

  /*for recording using .wav file */
  if (wav) {
    rec_size_wav=0;
    enc_wav.channel=1;
    enc_wav.encoding=0x0007;
    strcpy(enc_wav.magic,"RIFF");
    enc_wav.sample_rate=8000;
    sprintf(fn_wav,"conf_%s.wav",id);
    if((fp_wav=CreateSoundFile(fn_wav,enc_wav))==NULL)
      debug("RTPSendThread","File %s cannot open\n", fn_wav);
  }

  /***** move rtp recording to RTPReceiveThread
  //for recording using .rtp file 
  sprintf(fn_rtp,"conf_%d.rtp",confid);
  if((fp_rtp=fopen(fn_rtp,"wb"))==NULL)
    debug("RTPSendThread","File %s cannot open\n",fn_rtp);
  ************/
 
 
  debug("RTPSendThread", "got the lock\n");

  outputsize =  default_duration * 8 * sizeof(unsigned short);
  mixed = NewBuffer(outputsize);
  indiv = NewBuffer(outputsize);
  safebufdata = NewBuffer(outputsize);  /*buffer to be written to file*/

  /* Current time */
  gettimeofday(&initialtime, NULL);
  total_duration = 0.0;
  duration_interval = (double) default_duration/1000.0;
  debug("RTPSendThread", "Duration Interval is %g\n",
        duration_interval);

  /* While thread is not terminated */

  while ( !conf->oTerminate ) {
    /* Wait for default_duration */
    next = timespec_delay(initialtime, total_duration);
    total_duration += duration_interval;
    /*
    debug("RTPSendThread", "Total Interval is %g\n",
        total_duration);
    */

    /* XXX check if the duration is negative or in past */
    COND_TIMEDWAIT(conf->cond, conf->mutex, next);
    if ( conf->oTerminate ) {
      break; /* from while loop */
    }

    gettimeofday(&tv, NULL);

    a2 = (tv.tv_sec * 1000 + 
          tv.tv_usec / 1000) * 8; /* sampling clock */
    a1 = a2 - default_duration * 8;

    memset(mixed->data, 0, outputsize);
    mixed->size = 0;

#ifdef INCLUDE_SSRC
    ssrc_count = 0;
#endif

    /* Mix the audio from all participants */
    for ( k=0; k<ListCount(&conf->participants); k++) {

      //debug("RTPSendThread", "Participant %d\n", k);

      userid = ListElement(&conf->participants, k);
      if ( !userid ) 
        continue;
      user   = ListElement(&glUsers, *userid); /* participant */
      if ( !user ) 
        continue;

      LOCK(user->mutex);

      oBreak = 0;
      user->oContributed = 0;
      user->tsinc += (default_duration * 8); /* Number of samples */

      for ( raw = user->rxbuflist; 
            raw && !oBreak ; raw = raw->next ) {
        //debug("RTPSendThread", "for loop ... %d, %d\n",raw->size, __LINE__);

        raw->oRemove = 0;

        b1 = raw->param;
        b2 = raw->param + raw->size / sizeof(unsigned short);

        //debug("RTPSendThread", "a1=%u,a2=%u,b1=%u,b2=%u\n", a1, a2, b1, b2);

        if ( b2 <= a1 ) {
          /* XXX remove the packet without mixing */
          raw->oRemove = 1;
          //printf("1\n");
        }
        else if ( b1 <= a1 && b2 > a1 && b2 <= a2 ) {
          for ( i = a1-b1, j=0; i < b2-b1; i++, j++) {
            ((short *) mixed->data)[j] += 
              ((short *) raw->data)[i];
          }
          raw->oRemove = 1;
          user->oContributed = 1;
          //printf("2\n");
        }
        else if ( b1 >= a1 && b2 <= a2 ) {
          for ( i = 0, j=b1-a1; i< b2-b1; i++, j++) {
            (( short *) mixed->data)[j] += 
              (( short *) raw->data)[i];
          }
          raw->oRemove = 1;
          user->oContributed = 1;
          //printf("3\n");
        }
        else if ( b1 > a1 && b1 < a2 && b2 >= a2 ) {
          for ( i = 0, j=b1-a1; i< a2-b1; i++, j++) {
            (( short *) mixed->data)[j] += 
              (( short *) raw->data)[i];
          }
          user->oContributed = 1;
          //printf("4\n");
        }
        else if ( b1 <= a1 && b2 >= a2 ) {
          for ( i = a1 - b1, j=0; j< a2-a1; i++, j++) {
            (( short *) mixed->data)[j] += 
              (( short *) raw->data)[i];
          }
          user->oContributed = 1;
          //printf("5\n");
        }
        else if ( b1 >= a2 ) {
          oBreak = 1;
          /* Do nothing */
          //printf("6\n");
        }
        else {
          debug("RTPSendThread", "Error, should not come here\n");
        }

      }

      if ( user->oContributed ) {
        /* XXX add it into CSRC list */
        mixed->size = outputsize;

#ifdef INCLUDE_SSRC
        if ( ssrc_count < sizeof(ssrc_list) - 1 ) {
          ssrc_list[ssrc_count] = user->ssrc;
          ssrc_count++;
        }
        //debug("RTPMSendThread", "contributor ssrc %u\n", user->ssrc);
#endif
      }
      
      UNLOCK(user->mutex);
    }

#if 0 /* Local playback */
    {
      static int au = -1;
      static long bytescount = 0;
      static long timecount = 0;
      
      if ( au == -1 ) {
        au = open("/dev/audio", O_RDWR);
      }
      //debug("Playing", "%d bytes\n", encoded->size);
      write(au, encoded->data, encoded->size);
      
      bytescount += encoded->size;
      timecount += default_duration;
      
      if ( timecount >= 3000 ) {
        /* 3 seconds */
        debug("RTPSendThread", "bandwidth=%g kb/s\n",
              (bytescount * 8.0) / timecount);
        timecount = 0;
        bytescount = 0;
      }
    }
#endif


/* we record the conf into files here */
 if(mixed->size > 0) {
   if(au) {
     /*write to .au file first */
     /*search the point inside file to write data */
     f_offset = ftell(fp_au);
     fseek(fp_au, f_offset, SEEK_SET);
     /* perform transcoding */
     safebufdata = MediaTranscode("l16","pcmu/8000", mixed->data, outputsize);
     if(safebufdata == NULL)
       debug("RTPSendThread","Error: au safebufdata MediaTranscode\n");
     fwrite(safebufdata->data, sizeof(char), safebufdata->size, fp_au);
     rec_size_au = rec_size_au+safebufdata->size;
   }

   if (wav) {
     /*write to .wav file second */
     f_offset = ftell(fp_wav);
     fseek(fp_wav, f_offset, SEEK_SET);
     /*perform transcoding; need confirming following encode type!! */
     safebufdata = MediaTranscode("l16","pcmu/8000", mixed->data, outputsize);
     if(safebufdata==NULL)
       debug("RTPSendThread","Error: wav safebufdata MediaTranscode\n");
     fwrite(safebufdata->data, sizeof(char), safebufdata->size, fp_wav);
     rec_size_wav+=safebufdata->size;
   }

 } /* end if(mixed->size >0) */

    /* Update the CSRC list */

    {
      //debug("RTPSendThread", "Sending packet\n");
#ifdef INCLUDE_SSRC
      int idx1, idx2, idx3;
      unsigned int ssrc1, ssrc2;
#endif
      
      for ( k=0; k<ListCount(&conf->participants); k++) {
        int pt;
        
        userid = ListElement(&conf->participants, k);
        if ( !userid ) 
          continue;
        user   = ListElement(&glUsers, *userid); /* participant */
        if ( !user ) 
          continue;
        LOCK(user->mutex);

#ifdef INCLUDE_SSRC

      /* adding to CSRC */
      for (idx1 =0; idx1 < ssrc_count; idx1 ++ ) {
        int oFound = 0;
        
        ssrc1 = ssrc_list[idx1];
        if ( ssrc1 == user->ssrc ) {
          continue;
        }
        for ( idx2=0; idx2 < last_ssrc_count; idx2++) {
          ssrc2 = last_ssrc_list[idx2];
          if ( ssrc1 == ssrc2 ) {
            oFound = 1;
            break;
          }
        }
        if ( !oFound ) {
          debug("RTPSendThread", "Adding %u\n", ssrc1);
          the_err = RTPSessionAddToCSRCList (user->rtp_context,
                                             ssrc1);
          if ( the_err != RTP_OK ) {
            debug("RTPSessionAddToCSRCList", "%s\n",
                  RTPDebugStr());
          }
        }
      }


      /* removing from CSRC */
      for (idx2 =0; idx2 < last_ssrc_count; idx2 ++ ) {
        int oFound = 0;
        
        ssrc2 = last_ssrc_list[idx2];
        if ( ssrc2 == user->ssrc ) {
          continue;
        }
        for ( idx1=0; idx1 < ssrc_count; idx1++) {
          ssrc1 = ssrc_list[idx1];
          if ( ssrc1 == ssrc2 ) {
            oFound = 1;
            break;
          }
        }
        if ( !oFound ) {
          debug("RTPMSendThread", "Removing %u, self is %u\n", ssrc2,
                user->ssrc);
 

          the_err = RTPSessionRemoveFromCSRCList (user->rtp_context,
                                                  ssrc2);
          if ( the_err != RTP_OK ) {
            debug("RTPSessionRemoveFromCSRCList", "%s\n",
                  RTPDebugStr());
          }
        }
      }

#endif

      if ( mixed->size > 0 ) {
        /* 
         * Send mixed stream to each participant.
         * A participant should not get his own audio.
         */
        pt = -1;
        ptstr = GetPayloadType(&pt, &user->sdp_tx);
        if ( pt == -1 || ptstr == NULL ) {
          /* Ignore the error */
        }
        else {        
          int oSilence = 1;

          if ( !oSendBackAudio && user->oContributed ) {
            /*
            debug("RTPSendThread", "user %d, %d has contributed\n",
                  *userid, k);
            */

            /* substract the user stream */
            memcpy(indiv->data, mixed->data, outputsize);

            oBreak = 0;
            for ( raw = user->rxbuflist; 
                  raw && !oBreak ; raw = raw->next) {

              b1 = raw->param;
              b2 = raw->param + raw->size / sizeof(unsigned short);

              if ( b2 <= a1 ) {
              }
              else if ( b1 <= a1 && b2 > a1 && b2 <= a2 ) {
                for ( i = a1-b1, j=0; i < b2-b1; i++, j++) {
                  ((short *) indiv->data)[j] -= 
                    ((short *) raw->data)[i];
                }
              }
              else if ( b1 >= a1 && b2 <= a2 ) {
                for ( i = 0, j=b1-a1; i< b2-b1; i++, j++) {
                  (( short *) indiv->data)[j] -= 
                    (( short *) raw->data)[i];
                }
              }
              else if ( b1 > a1 && b1 < a2 && b2 >= a2 ) {
                for ( i = 0, j=b1-a1; i< a2-b1; i++, j++) {
                  (( short *) indiv->data)[j] -= 
                    (( short *) raw->data)[i];
                }
              }
              else if ( b1 <= a1 && b2 >= a2 ) {
                for ( i = a1 - b1, j=0; j< a2-a1; i++, j++) {
                  (( short *) indiv->data)[j] -= 
                    (( short *) raw->data)[i];
                }
              }
              else if ( b1 >= a2 ) {
                oBreak = 1;
              }
              else {
                debug("RTPSendThread", "Error, should not come here\n");
              }
              
            }

            oSilence = 1;
            for ( i=0; i<outputsize/2; i++) {
              if ( ((short *)indiv->data)[i] != 0 ) {
                oSilence = 0;
                break;
              }
            }
            if ( !oSilence ) {
              encoded = MediaTranscode("l16", ptstr, indiv->data,
                                   outputsize);
            }
            else {
              encoded = NULL;
            }
          }
          else {
            /* Send the mixed stream */
            encoded = MediaTranscode("l16", ptstr, mixed->data,
                                   outputsize);
          }

          if ( encoded != NULL ) {
#ifdef INCLUDE_SSRC //#if 0
            for ( idx1 = 0; idx1 < ssrc_count; idx1++ ) {
              ssrc1 = ssrc_list[idx1];
              if ( ssrc1 == user->ssrc )
                continue;

              the_err = RTPSessionAddToCSRCList (user->rtp_context,
                                                 ssrc1);
              if ( the_err != RTP_OK ) {
                debug("RTPSessionAddToCSRCList", "%s\n",
                      RTPDebugStr());
              }
              
            }
#endif

            /* Send the RTP packet */

            //debug("RTPSendThread", "Calling RTPSend() %d bytes\n", encoded->size);
            the_err = RTPSend(user->rtp_context, 
                              (int32) user->tsinc,
                              (int8) marker,
                              (int16) pt, 
                              (int8 *) encoded->data,
                              (int) encoded->size);
            
            if ( the_err != RTP_OK ) {
              static int fail_count = 0;
              if ( fail_count == 0 ) { 
                debug("RTPSendThread", "context(%d), RTPSend %s\n", 
                      user->rtp_context,
                      RTPDebugStr());
              }
              fail_count ++;
              /* Count only one in 29 packets as failed */
              if ( fail_count >= 29 ) {
                fail_count = 0;
              } 
            }
            else {
              //debug("RTPSendThread", "Sending %d bytes\n", encoded->size);
              /*
              if ( user->tsinc != default_duration * 8 ) {
                debug("RTPSend", "called with tsinc = %d\n", user->tsinc);
              }
              */
            }

#ifdef INCLUDE_SSRC //#if 0
            for ( idx1 = 0; idx1 < ssrc_count; idx1++ ) {
              ssrc1 = ssrc_list[idx1];
              if ( ssrc1 == user->ssrc )
                continue;
              the_err = RTPSessionRemoveFromCSRCList (user->rtp_context,
                                                      ssrc1);
              if ( the_err != RTP_OK ) {
                debug("RTPSessionRemoveFromCSRCList", "%s\n",
                      RTPDebugStr());
              }
            }
#endif

            user->tsinc = 0; /* set this to 0 */

#if 1 /* Bitrate display */
            {
              bytescount += encoded->size;
              timecount += default_duration;
              
              if ( actualtime == 0 ) {
                actualtime = time(0);
              }
              
              /* Print bandwidth every 5 seconds */
              if ( timecount >= 5000 ) {
                /* 3 seconds */
                long interv = time(0) - actualtime;
                debug("RTPSendThread", "expected=%g kb/s, actual=%g kb/s\n",
                      (bytescount * 8.0) / timecount,
                      interv <= 0 ? 0 : 
                      (bytescount * 8.0) / (interv * 1000));
                timecount = 0;
                bytescount = 0;
                actualtime = time(0);
              }
            }
#endif

            FreeBuffer(encoded);
          }
        }
        } /* mixed->size > 0 */
        
        UNLOCK(user->mutex);
      }
    }

    /* Remove old buffers */

    for ( k=0; k<ListCount(&conf->participants); k++) {

      //debug("RTPSendThread", "Participant %d\n", k);

      userid = ListElement(&conf->participants, k);
      if ( !userid ) 
        continue;
      user   = ListElement(&glUsers, *userid); /* participant */
      if ( !user ) 
        continue;
      LOCK(user->mutex);

      oBreak = 0;
      for ( raw = user->rxbuflist, prevptr = NULL; 
            raw && !oBreak ; ) {
        /*
         * XXX remove should be done in other loop to avoid
         * getting own audio 
         */

        b1 = raw->param;
        b2 = raw->param + raw->size / sizeof(unsigned short);

        if ( b1 >= a2 ) {
          oBreak = 1;
        }

        if ( raw->oRemove ) {
          if ( prevptr == NULL ) {
            /* First element in the list */
            user->rxbuflist = user->rxbuflist->next;
            FreeBuffer(raw);
            raw = user->rxbuflist;
          }
          else {
            prevptr->next = prevptr->next->next;
            FreeBuffer(raw);
            raw = prevptr->next;
          }
          //debug("RTPSendThread", "deleted\n");
        }
        else {
          prevptr = raw;
          raw = raw->next;
        }

      } /* end of for all packets */
      UNLOCK(user->mutex);
    }

    if ( mixed->size > 0 && marker == 1) {
      marker = 0; /* begining of talk spurt */
    }
    else if ( marker == 0 && mixed->size == 0 ) {
      marker = 1;
    }

#ifdef INCLUDE_SSRC
    last_ssrc_count = ssrc_count;
    //debug("RTPMSendThread", "current->last %d\n", ssrc_count);
    for ( idx3 = 0; idx3 < ssrc_count; idx3 ++ ) {
      last_ssrc_list[idx3] = ssrc_list[idx3];
    }
    ssrc_count = 0;
#endif
  }

  FreeBuffer(indiv);
  FreeBuffer(mixed);
  FreeBuffer(safebufdata);

  if (au)
    CloseSoundFile(fp_au,rec_size_au);
  if (wav)
    CloseSoundFile(fp_wav,rec_size_wav);

  UNLOCK(conf->mutex);

  return NULL;
}

/* Set the flag to send back the audio to the participant
 */
void RTPMSendBackAudioToParticipant(int true_or_false)
{
  oSendBackAudio = true_or_false;
}












