# Columbia server makefile fragment for rtspd
# Copyright 1998-2000 by Columbia University; all rights reserved 

rtspd_dir := rtspd

RTSPD_MODULES := docs

# Include the description for each sub-module
include $(patsubst %,$(topsrcdir)/$(rtspd_dir)/%/module.mk,$(RTSPD_MODULES))


RTSPD_BIN := \
  $(rtspd_dir)/rtspd

RTSPD_SRC = \
  $(rtspd_dir)/config.c \
  $(rtspd_dir)/announce.c \
  $(rtspd_dir)/delete.c \
  $(rtspd_dir)/describe.c \
  $(rtspd_dir)/getusername.c \
  $(rtspd_dir)/mime.c \
  $(rtspd_dir)/pause.c \
  $(rtspd_dir)/play.c \
  $(rtspd_dir)/playservice.c \
  $(rtspd_dir)/qt.c \
  $(rtspd_dir)/recordservice.c \
  $(rtspd_dir)/request.c \
  $(rtspd_dir)/response.c \
  $(rtspd_dir)/rtsp.c \
  $(rtspd_dir)/rtspd.c \
  $(rtspd_dir)/service.c \
  $(rtspd_dir)/session.c \
  $(rtspd_dir)/setup.c \
  $(rtspd_dir)/snd.c \
  $(rtspd_dir)/stream.c \
  $(rtspd_dir)/teardown.c \
  $(rtspd_dir)/mp3.c \
  $(rtspd_dir)/rf.c \
  $(rtspd_dir)/rtpfile.c \
  $(rtspd_dir)/wav.c \
  $(rtspd_dir)/mpv.cpp \
  $(rtspd_dir)/mparse.cpp

RTSPD_MAYBESRC := \
  $(rtspd_dir)/authen.c \
  $(rtspd_dir)/authen_test.c \
  $(rtspd_dir)/parser_test.c \
  $(rtspd_dir)/realmedia.c


RTSPD_HEADERS := \
  $(rtspd_dir)/audio.h \
  $(rtspd_dir)/authen.h \
  $(rtspd_dir)/config.h \
  $(rtspd_dir)/getusername.h \
  $(rtspd_dir)/qt.h \
  $(rtspd_dir)/queue.h \
  $(rtspd_dir)/realmedia.h \
  $(rtspd_dir)/request-rtsp.h \
  $(rtspd_dir)/rtp.h \
  $(rtspd_dir)/rtp_api.h \
  $(rtspd_dir)/rtp_pt.h \
  $(rtspd_dir)/rtsp.h \
  $(rtspd_dir)/rtspd.h \
  $(rtspd_dir)/snd.h \
  $(rtspd_dir)/stream.h \
  $(rtspd_dir)/wav.h  \
  $(rtspd_dir)/mime.h \
  $(rtspd_dir)/mp3.h \
  $(rtspd_dir)/rf.h \
  $(rtspd_dir)/rtpfile.h \
  $(rtspd_dir)/mpv.h \
  $(rtspd_dir)/mparse.h

RTSPD_MISC := \
  $(rtspd_dir)/module.mk \
  $(rtspd_dir)/rtspd.dsp

RTSPD_MISCDIST := \
  $(rtspd_dir)/TODO \
  $(rtspd_dir)/BUGS \
  $(rtspd_dir)/README \
  $(rtspd_dir)/mime.types

RTSPD_NEEDED_LIBRARIES_LIST := \
  $(rtspd_dir)/rtspd-needed-shared-libraries

RTSPD_NEEDED_LIBRARIES_SOURCES := \
  $(rtspd_dir)/rtspd-needed-shared-libraries-sources

RTSPD_OBJ_C := \
	$(patsubst %.c,%.o,$(filter %.c,$(RTSPD_SRC)))

RTSPD_OBJ_CPP := \
	$(patsubst %.cpp,%.o,$(filter %.cpp,$(RTSPD_SRC)))

RTSPD_OBJ := $(RTSPD_OBJ_C) $(RTSPD_OBJ_CPP)

RTSPD_DEP_C := \
	$(RTSPD_OBJ_C:.o=.d)

RTSPD_DEP_CPP := \
	$(RTSPD_OBJ_CPP:.o=.d)

RTSPD_DEP := $(RTSPD_DEP_C) $(RTSPD_DEP_CPP)

RTSPD_MISCCLEAN := \
	$(RTSPD_NEEDED_LIBRARIES_LIST)

# Include directories and compiler directives required locally
$(rtspd_dir)-CFLAGS := $(LIBCINE_INC) $(LIBDICT_INC) -DYULIYA # XXX

RTSPD_LIBS := $(LIBCINE_LIB) $(LIBDICT_LIB) \
	$(NISLIBS) $(SOCKLIBS)
RTSPD_DEPENDENCIES := $(LIBCINE) $(LIBDICT)
RTSPD_REQS := 
RTSPD_INC := -I$(topsrcdir)/$(rtspd_dir)

# all: $(rtspd_dir)/rtspd

# Note: we must link with $(CXX) (the C++ compiler) since rtspd includes some
# C++ code.
$(rtspd_dir)/rtspd:  $(RTSPD_OBJ) $(RTSPD_CPPOBJ) $(RTSPD_DEPENDENCIES)
	@echo "linking $@..."
	$(CXX) -o $@ $(RTSPD_OBJ) $(RTSPD_CPPOBJ) $(RTSPD_LIBS) $(LDFLAGS) $(LIBS)

$(RTSPD_NEEDED_LIBRARIES_SOURCES): $(rtspd_dir)/rtspd
	@echo "Building list of non-system shared libraries..."
	rm -f $(RTSPD_NEEDED_LIBRARIES_SOURCES)
	$(LDD) $(rtspd_dir)/rtspd | \
		sed -n -e '/not found/ s/^/XXX SET LD_LIBRARY_PATH: /p;' \
			-e '/=>/ s|^.* \([^ ]*/[^ ]*\).*$$|\1|p' | \
			egrep -v '^(/usr)?/lib/' \
			> $(RTSPD_NEEDED_LIBRARIES_SOURCES); \
			true
	if grep XXX $(RTSPD_NEEDED_LIBRARIES_SOURCES) > /dev/null; \
	then \
		echo "Missing shared library; can't continue"; \
		cat $(RTSPD_NEEDED_LIBRARIES_SOURCES); \
		rm -f $(RTSPD_NEEDED_LIBRARIES_SOURCES); \
		exit 1; \
	fi

$(RTSPD_NEEDED_LIBRARIES_LIST): $(RTSPD_NEEDED_LIBRARIES_SOURCES)
	@echo "Building local names of non-system shared libraries..."
	rm -f $(RTSPD_NEEDED_LIBRARIES_LIST)
	@rtspd_needed_libraries=`cat $(RTSPD_NEEDED_LIBRARIES_SOURCES)`; \
	for library in $$rtspd_needed_libraries; \
	do\
		echo $(rtspd_dir)/`basename $$library`; \
	done > $(RTSPD_NEEDED_LIBRARIES_LIST)

.PHONY: rtspd
all: rtspd
rtspd: $(RTSPD_REQS) $(RTSPD_BIN)

BIN       += $(RTSPD_BIN) $(RTSPD_OTHER_BIN)
SRC       += $(RTSPD_SRC) $(RTSPD_CPPSRC) $(RTSPD_OTHER_SRC)
HEADERS   += $(RTSPD_HEADERS)
MISC      += $(RTSPD_MISC) $(RTSPD_MAYBESRC)
MISCDIST  += $(RTSPD_MISCDIST)
MISCCLEAN += $(RTSPD_OTHER_MISCCLEAN)

RTSPD_DIST := $(RTSPD_SRC) $(RTSPD_CPPSRC) \
        $(RTSPD_MAYBESRC) $(RTSPD_HEADERS) \
	$(RTSPD_MISC) $(RTSPD_MISCDIST)

RTSPD_BINDIST := $(RTSPD_BIN) $(RTSPD_MISCDIST)

RTSPD_MISC_DLLS :=

RTSPD_BIN_WIN32 := \
  $(rtspd_dir)/Debug/rtspd.exe \
  $(rtspd_dir)/Release/rtspd.exe

RTSPD_BINDIST_WIN32 := $(RTSPD_MISCDIST) \
		$(RTSPD_BIN_WIN32) $(RTSPD_MISC_DLLS)

.PHONY: rtspd-dist rtspd-bindist rtspd-bindist-win32 copy-rtspd-shared-libraries

copy-rtspd-shared-libraries: $(RTSPD_NEEDED_LIBRARIES_SOURCES)
	@echo "Copying non-system shared libraries..."
	@rtspd_needed_libraries=`cat $(RTSPD_NEEDED_LIBRARIES_SOURCES)`; \
	if test "x$$rtspd_needed_libraries" != "x"; \
	then\
		echo cp $$rtspd_needed_libraries $(rtspd_dir); \
		cp $$rtspd_needed_libraries $(rtspd_dir); \
	fi


RTSPD_DIST_NAME := rtspd-$(VERSION)

RTSPD_DIST_FILES := \
		$(TOP_DIST) \
		$(LIBCINE_DIST) \
		$(LIBDICT_DIST) \
		$(RTSPD_DIST) \
		$(WIN32_DIST) \
		$(RTSPD_DOCS_DIST) \
		$(DOC_DIST) \
		$(EXTRA_DIST_FILES)

rtspd-dist: win32-dist
	$(MKDIST) $(topsrcdir) $(RTSPD_DIST_NAME) \
	$(RTSPD_DIST_FILES)

RTSPD_BINDIST_NAME := rtspd-$(VERSION)-$(SYSNAME)

RTSPD_BINDIST_FILES := \
		$(TOP_BINDIST) \
		$(LIBCINE_BINDIST) \
		$(RTSPD_BINDIST) \
		$(RTSPD_DOCS_BINDIST) \
		$(DOC_BINDIST) \
		`cat $(RTSPD_NEEDED_LIBRARIES_LIST)` \
		$(EXTRA_BINDIST_FILES)

rtspd-bindist: rtspd copy-rtspd-shared-libraries \
               $(RTSPD_NEEDED_LIBRARIES_LIST)
	$(MKDIST) $(topsrcdir) $(RTSPD_BINDIST_NAME) \
	$(RTSPD_BINDIST_FILES)

RTSPD_BINDIST_WIN32_NAME := rtspd-$(VERSION)-$(DATE)-Win32

RTSPD_BINDIST_WIN32_FILES := \
		$(TOP_BINDIST) \
		$(LIBCINE_BINDIST) \
		$(RTSPD_BINDIST_WIN32) \
		$(RTSPD_DOCS_BINDIST) \
		$(DOC_BINDIST) \
		$(EXTRA_BINDIST_FILES)

rtspd-bindist-win32:
	$(MKDIST) $(topsrcdir) $(RTSPD_BINDIST_WIN32_NAME) \
	$(RTSPD_BINDIST_WIN32_FILES)

rtspd-help: 
	@echo "RTSPD (Media Server) Make Targets"
	@echo " make rtspd                Make rtspd executable on Unix."
	@echo " make rtspd-dist           Make rtspd source distribution."
	@echo " make rtspd-bindist        Make rtspd binary distribution for Unix."
	@echo " make rtspd-bindist-win32  Make rtspd binary distribution for Win32."

help: rtspd-help
